/*---------------------------------------------------------------------------*\

	FILE....: measerl.cpp
	TYPE....: C++ Program
	AUTHOR..: David Rowe
	DATE....: 3/5/01

	Test program to measure the Echo Return Loss (ERL) of a VPB port.
	The idea is to minimise the ERL (maximum -ve number).

	Instructions:
	1. ./measerl --port [1..4]
	2. Dial the line connected to the port being tested
	3. A series of tones will be heard in the phone dialling the port,
	   avoid speaking into the phone to prevent corrupting the results.
	4. The program will print the results.

\*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

         Voicetronix Voice Processing Board (VPB) Software

         Copyright (C) 1999-2001 Voicetronix www.voicetronix.com.au

         This library is free software; you can redistribute it and/or
         modify it under the terms of the GNU Lesser General Public
         License as published by the Free Software Foundation; either
         version 2.1 of the License, or (at your option) any later version.

         This library is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty of
         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
         License along with this library; if not, write to the Free Software
         Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	 USA

\*---------------------------------------------------------------------------*/

#include "../src/mess.h"
#include "../src/comm.h"
#include "../src/config.h"
#include "../src/dspfifo.h"
#include "../src/timer.h"
#include "../src/timer.h"
#include "../src/wobbly.h"
#include "../src/translate.h"
#include "../src/mapdev.h"
#include "../src/vpbapi.h"
#include "../src/verbose.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define WAIT		1	// time out delay 
#define	BOARD		0	// board number
#define	N	        160	// size of processing frame
#define OFF_HK_DELAY    500     // number of ms to pause after off-hook
#define NSAM            40      // number of frames to sample
#define NCONV           20      // number of frames to wait for convergence
#define AMP             4000    // peak amplitude of test signal
#define NFREQS          5       // number of test frequencies 
#define FS              8000.0  // sampling frequency

float freqs[] = {300.0, 500.0, 1000.0, 2000.0, 3000.0};

typedef struct {
	int bal_min, bal_max, reg;
} SCAN_LUT;

SCAN_LUT scan_lut[] = {
	{192, 200, 0x32},
	{  0, 255, 0x3a},
	{  0, 255, 0x42}
};

int scan;
FILE *fscan;

void vpb_set_codec_reg(int chdev,    // channel handle
		       unsigned short addr,  // 8-bit address of codec register
		       unsigned short data   // 8-bit data to write to register
		      );

void play_set_gain(int handle, float gain, Comm *c);
void record_set_gain(int handle, float gain, Comm *c);

void scan_erl(int h, float *av_erl, float *min_erl);
float measure_erl(int h, float f);
int arg_exists(int argc, char *argv[], char *arg);

// headers for functions internal to driver
int vpb_echo_canc_disable();
int vpb_echo_canc_enable();

int kbhit();

int main(int argc, char *argv[])
{
	int             arg,h,hdial;
	int bal, bal2, bal3, port, on_hook, dport;
	float av_erl, playg, recg, max_erl;
	VPB_EVENT e;
	char s[MAX_STR];
	int bal_min, bal_max, bal_step; 
	int best_bal1, best_bal3;

	// scan command line arguments

	// help
	if (arg_exists(argc, argv, "-h") || arg_exists(argc, argv, "--help")) {
		printf("--bal1 [Hex]        Set codec hybrid balance 1 reg\n");
		printf("--bal2 [Hex]        Set codec hybrid balance 2 reg\n");
		printf("--bal3 [Hex]        Set codec hybrid balance 3 reg\n");
		printf("--port [1..4]       VPB port\n");
		printf("--ec                Enable echo canceller\n");
		printf("--pg [-12.0..12.0]  Set codec play gain\n");
		printf("--rg [-12.0..12.0]  Set codec record gain\n");
		printf("--dial dport number Dial out another port\n");
		printf("--offhook          Take off hook immediately\n");
		printf("--scan [1..3] TextFile.txt start[0..ff] end[0..ff] step[e.g. 1]\n");
		printf("--autoscan         Auto scan codec balance values\n");
		printf("--handset          proto handset mode\n");
		printf("--loop             Enable software loopback\n");
		printf("--hostecho         Enable Host Echo Canceller\n");

	        exit(0);
	}

	// option to change channel
	if ((arg = arg_exists(argc, argv, "--port")) != 0) {
		port = atoi(argv[arg+1]);
		assert((port >= 1) && (port <= 12));
	}
	else 
		port = 1;

	// initialise 

	try {
	        //verbose(1);		
		h = vpb_open(1,port);

		// option to turn on loop back
		if ((arg = arg_exists(argc, argv, "--loop")) != 0) 
			vpb_loopback_on(h);
		
		// option to turn on host echo
		if ((arg = arg_exists(argc, argv, "--hostecho")) != 0) 
			vpb_hostecho_on(h);

		// option to set codec balance reg 1
		bal = -1;
		if ((arg = arg_exists(argc, argv, "--bal1")) != 0) {
			bal = strtol(argv[arg+1], NULL, 16);
			assert((bal >= 0) && (bal <= 255));
			vpb_set_codec_reg(h, 0x32, bal);
			printf("bal1 = 0x%02x (%d)\n", bal, bal);
		}

		// option to set codec balance reg 2
		bal2 = -1;
		if ((arg = arg_exists(argc, argv, "--bal2")) != 0) {
			bal2 = strtol(argv[arg+1], NULL, 16);
			assert((bal2 >= 0) && (bal2 <= 255));
			vpb_set_codec_reg(h, 0x3a, bal2);
			printf("bal2 = 0x%02x (%d)\n", bal2, bal2);
	        }

		// option to set codec balance reg 3
		bal3 = -1;
		if ((arg = arg_exists(argc, argv, "--bal3")) != 0) {
			bal3 = strtol(argv[arg+1], NULL, 16);
			assert((bal3 >= 0) && (bal3 <= 255));

			if (!(bal & 32)) {
				printf("bit 5 of bal1 needs to be set to use bal3....\n");
				bal |= 32;
				vpb_set_codec_reg(h, 0x32, bal);
				printf("bal1 now set to 0x%02x\n", bal);
			}
			vpb_set_codec_reg(h, 0x42, bal3);
			printf("bal3 = 0x%02x (%d)\n", bal3, bal3);
		}

		// option to use OpenLine4 handset proto mod
		if ((arg = arg_exists(argc, argv, "--handset")) != 0) {
			// special mod for Simple Simon OpenLine4 hardware
			// which were specially modifed V16 OpenLine4's 
			// (V4PCIs)

			vpb_get_model(s);
			if (strcmp(s,"VPB4") == 0) {
				printf("manope\n");
				// audio sw but not loop
				vpb_set_codec_reg(h, 0x12, 0x7c);
				vpb_set_codec_reg(h, 0x0a, 0x20); 
			}
		
		}

		// option to set codec record gain
		if ((arg = arg_exists(argc, argv, "--rg")) != 0) {
			recg = atof(argv[arg+1]);
			assert((recg >= -12.0) && (recg <= 12.0));
			vpb_record_set_hw_gain(h, recg);
			printf("record hw gain set to %3.2f dB\n", recg);
		} 

		// option to set codec play gain
		if ((arg = arg_exists(argc, argv, "--pg")) != 0) {
			playg = atof(argv[arg+1]);
			assert((playg >= -12.0) && (playg <= 12.0));
			vpb_play_set_hw_gain(h, playg);
			printf("play hw gain set to %3.2f dB\n", playg);
		}

		// option to use echo canceller (off by default)
		if (!arg_exists(argc, argv, "--ec")) {
			vpb_echo_canc_disable();
		}

		// option to dial out of another port
		if ((arg = arg_exists(argc, argv, "--dial")) != 0) {
			dport = atoi(argv[arg+1]);
			assert((dport >= 1) && (dport <= 4));
			assert(dport != port);
			hdial = vpb_open(1, dport);
			vpb_sethook_sync(hdial, VPB_OFFHOOK);
			printf("OK, dialing %s down port %d\n", argv[arg+2],
			       dport);
			vpb_dial_sync(hdial, argv[arg+2]);
		}
		
		// option to enter scan mode 
		arg = arg_exists(argc, argv, "--scan");
		if (arg) {
			scan = atoi(argv[arg+1]);
			assert((scan == 1) || (scan == 2) || (scan == 3));
			fscan = fopen(argv[arg+2],"wt");
			assert(fscan != NULL);

			bal_min = strtol(argv[arg+3], NULL, 16);
			assert((bal_min >= 0) && (bal_min <= 255));
			bal_max = strtol(argv[arg+4], NULL, 16);
			assert((bal_max >= 0) && (bal_max <= 255));
			bal_step = strtol(argv[arg+5], NULL, 16);
			assert((bal_step >= 0) && (bal_step <= 255));
		}
		else
			scan = 0;

		// start main processing loop ---------------------------

		// wait for ring and take off hook

		if ((arg = arg_exists(argc, argv, "--dial")) != 0) {
			printf("Waiting for rings on port %d\n",port);
		} 
		else {
			if (!arg_exists(argc, argv, "--offhook")) {
				printf("Please dial line "
				       "attached to port %d\n",port);
			}
		}

		// see if we should take off hook immediately

		if (arg_exists(argc, argv, "--offhook")) {
			on_hook = 0;
			vpb_sethook_sync(h, VPB_OFFHOOK);
			vpb_sleep(300);
		}
		else {
		  on_hook = 1;

	 	  // see if we should proceed without changing hook switch
		  if (arg_exists(argc, argv, "--nohook")) {
		    on_hook = 0;
		  }
		  else 
		    on_hook = 1;
		}

		do {
			while(vpb_get_event_async(&e) == VPB_OK) {
				vpb_translate_event(&e, s);
				mprintf(s);

				// take channels off-hook on ring
				if (e.type == VPB_RING) {
					vpb_sethook_sync(e.handle,VPB_OFFHOOK);
					on_hook = 0;
				}
			}
			vpb_sleep(100);
		} while(on_hook);

		if (scan) fprintf(fscan, "scan = %d bal = 0x%x bal2 = 0x%x"
				  "bal3 = 0x%x\n", scan, bal, bal2, bal3);
		printf(" bal    300    500   1000   2000   3000"
		       "     av    max  best_b  best_max");
		if (scan) fprintf(fscan, " bal    300    500   1000   2000   3000"
				  "     av    max  best_b  best_max");
		int best_bal, b, reg;
		float best_max_erl, best_av_erl;

		if (scan || arg_exists(argc, argv, "--autoscan")) {

			if (arg_exists(argc, argv, "--autoscan")) {
				fscan = fopen("autoscan.txt","wt");
				assert(fscan != NULL);

				// first scan bal1
				scan = 1;
				bal_min = 192;
				bal_max = 216;
				bal_step = 1;

			}

			if ((scan==3) && !(bal & 32)) {
				printf("\nbit 5 of bal1 needs to be set to use bal3....\n");
				bal |= 32;
				vpb_set_codec_reg(h, 0x32, bal);
				printf("bal1 now set to 0x%02x\n", bal);
			}
		        reg = scan_lut[scan-1].reg;	
			best_max_erl = 100.0;
			best_bal = 0;

			for(b=bal_min; (b<=bal_max) && !kbhit(); b+=bal_step) {
				printf("\n0x%02x ",b);
				if (scan) fprintf(fscan, "\n0x%0x ",b);
				vpb_set_codec_reg(h, reg, b);
				scan_erl(h, &av_erl, &max_erl);
				if (max_erl < best_max_erl) {
					best_bal = b;
					best_max_erl = max_erl;
					best_av_erl = av_erl;
				}
				printf("   0x%02x %9.2f ",best_bal,best_max_erl);
				if (scan) fprintf(fscan, ")   x%02x %9.2f ", 
						  best_bal, best_max_erl);

			}

			if (arg_exists(argc, argv, "--autoscan")) {
				// set bal12 to best value & enable bal3
				best_bal1 = best_bal |= 32;
				vpb_set_codec_reg(h, 0x32, best_bal1);
				printf("\n\nbal1 = 0x%02x (%d)\n", best_bal1, best_bal1);

				// now scan bal3
				scan = 3;
				bal_min = 0;
				bal_max = 255;
				bal_step = 1;

				reg = scan_lut[scan-1].reg;	
				best_max_erl = 100.0;
				best_bal = 0;

				printf("Scanning balance register 3...........\n");

				for(b=bal_min; (b<=bal_max) && !kbhit(); b+=bal_step) {
					printf("\n0x%02x ",b);
					if (scan) fprintf(fscan, "\n0x%02x ",b);
					vpb_set_codec_reg(h, reg, b);
					scan_erl(h, &av_erl, &max_erl);
					if (max_erl < best_max_erl) {
						best_bal = b;
						best_max_erl = max_erl;
						best_av_erl = av_erl;
					}
					printf("   0x%02x %9.2f ",best_bal,best_max_erl);
					if (scan) fprintf(fscan, ")   x%02x %9.2f ", 
							  best_bal, best_max_erl);

				}
				best_bal3 = best_bal;
			}

		} 
		else {
			// single mesurement
			printf("\n     ");
			scan_erl(h, &av_erl, &max_erl);
		}

		if (scan) fclose(fscan);
		printf("\n");

		if ((arg = arg_exists(argc, argv, "--loop")) != 0) 
			vpb_loopback_off(h);

		if ((arg = arg_exists(argc, argv, "--hostecho")) != 0) 
			vpb_hostecho_off(h);

		vpb_sethook_sync(h, VPB_ONHOOK);
		vpb_close(h);
		if ((arg = arg_exists(argc, argv, "--dial")) != 0) {
			vpb_sethook_sync(hdial, VPB_ONHOOK);
			vpb_close(hdial);
		}

		if (arg_exists(argc, argv, "--autoscan")) {
			printf("Autoscan run finished!\n\n");
			printf("set bal1 = 0x%02x\n", best_bal1);
			printf("set bal3 = 0x%02x\n", best_bal3);
			printf("\n");
			printf("e.g. using environment variables:\n");
			printf("export VPB_BAL1 = 0x%02x\n",best_bal1);
			printf("export VPB_BAL3 = 0x%02x\n",best_bal3);
			printf("\n");
			printf("Or using C code in your program:\n");
			printf("vpb_set_codec_reg(h, 0x32, 0x%02x);\n", best_bal1);
			printf("vpb_set_codec_reg(h, 0x42, 0x%02x);\n", best_bal3);
			printf("\n");
			printf("You can safely increase the total gains by: %3.2f dB\n",
			       best_max_erl+6.0);
			printf("e.g. using environment variables:\n");
		}
	}	// try ...

	catch (Wobbly w) {
		char	s[MAX_STR];
		
		w.translate(s);
		
		if (w.file[0] != 0) 
			printf("exception caught: %s, %s, line = %d\n"
			       ,s, w.file, w.line);
		else
			printf("exception caught: %s\n",s);

		printf("Press any key to exit....\n");
	}

	return 0;
}

// measeure erl across a range of frequencies

void scan_erl(int h, float *av_erl, float *max_erl) {
	float erl;
	int i;

	vpb_record_buf_start(h, VPB_LINEAR);
	vpb_play_buf_start(h, VPB_LINEAR);
	
	*av_erl = 0.0;
	*max_erl = -100.0;
	for(i=0; i<NFREQS; i++) {
		erl = measure_erl(h, freqs[i]);
		*av_erl += erl;
		if (erl > *max_erl)
			*max_erl = erl;
	}
	*av_erl /= NFREQS;

	// clean up and finish
	vpb_record_buf_finish(h);
	vpb_play_buf_finish(h);

	printf("%6.2f %6.2f ", *av_erl, *max_erl);
	if (scan) fprintf(fscan, "%6.2f %6.2f ", *av_erl, *max_erl);
}

// measure erl at a single frequency

float measure_erl(int h, float f) {
	int		i,j;
	float           sam;
        short           bufrd[N], bufwr[N];
	float           echo_pwr, ref_pwr;
	float           erl;
	float           pi = 4.0*atan(1.0);
	int             ne, nr;

	ref_pwr = echo_pwr = 0.0;
	ne = nr = 0;

	for(i=0; i<NSAM; i++) {
		  
		// generate a buffer of sine samples and output

		for(j=0; j<N; j++) { 
			  
			sam = cos(2*pi*(f/FS)*(i*N+j));
			sam *= AMP;

			bufwr[j] = (short)sam;
			if (i >= NCONV) {
				ref_pwr += sam*sam;
				nr++;
			}
		} 

		vpb_play_buf_sync(h, (char*)bufwr, sizeof(short)*N);

		// wait for buffer of echo samples

		vpb_record_buf_sync(h, (char*)bufrd, sizeof(short)*N);

		for(j=0; j<N; j++) { 
			sam = bufrd[j];
			if (i >= NCONV) {
				echo_pwr += sam*sam;
				ne++;
			}
		}
			  
	}

	// print results
		
	erl = 10*log10(echo_pwr/ref_pwr);
	printf("%6.2f ",erl);
	if (scan) fprintf(fscan, "%6.2f ",erl);
	return erl;
}

int arg_exists(int argc, char *argv[], char *arg) {
  int i;

  for(i=0; i<argc; i++)
    if (strcmp(argv[i],arg) == 0)
      return i;

  return 0;
}






